Synth's `string` generator type regroups many different useful functionalities, from the ability to generate sentences
from a regular expression to dates and times, through leveraging [Faker][faker] providers to generate valid addresses,
telephone numbers, etc.

## pattern

String values generated by a specified regular expression in the `pattern` key.

#### Example

```json synth
{        
  "type": "string",
  "pattern": "I am a (man|woman) with [2-9] friends"
}
```

## uuid

`uuid` generates hyphenated [UUIDs](https://en.wikipedia.org/wiki/Universally_unique_identifier).

This generator has no parameters.

#### Example

```json synth
{        
  "type": "string",
  "uuid": {}
}
```

## faker

Synth integrates with the Python [Faker library][faker]. To generate a string using `Faker`, use the `"faker": {...}`
key/value pair. The value of the `faker` key must be an object with at least a `generator` key.

#### Example

```json synth
{
  "type": "string",
  "faker": {
    "generator": "sentence"
  }
}
```

Faker offers a variety of different fake data, in different locales and can be leveraged easily from Synth schemas using
the `generator` key. Any of the Faker [standard providers](https://faker.readthedocs.io/en/master/providers.html)
generating a string can be used in a Synth schema. The example given below generates fake credit card numbers using
the [`credit_card_number`](https://faker.readthedocs.io/en/master/providers/faker.providers.credit_card.html#faker.providers.credit_card.Provider.credit_card_number)
provider.

Some Faker providers allow for parameters to customize the generated data. These parameters can be specified alongside
the `generator` field as arbitrary key/value pairs. In the example below, we used `"card_type" : "amex"`.

Locales can be specified using the *optional* `locales` key. Allowed values are the same as locales specified
in [Faker's locales documentation](https://faker.readthedocs.io/en/master/locales.html). In the example below, we
use `"locales": ["en_GB"]`. Multiple locales can be specified and Synth will sample at random among them.

#### Example

```json synth
{
  "type": "string",
  "faker": {
    "generator": "credit_card_number",
    "card_type": "amex",
    "locales": ["en_GB"]
  }
}
```

## date_time

A `date_time` is a variant of the `string` generator type that generates values from a specified range of `date_time`
values. This lets you, for example, generate valid days of the year for an `updated_at` column or a
valid [RFC 2822](https://tools.ietf.org/html/rfc2822) timestamp for an email header field, among others.

You can specify a string as a `date_time` by using the `"date_time": {...}` key/value pair.

#### Example

```json synth
{
  "type": "string",
  "date_time": {
    "format": "%Y-%m-%d",
    "subtype": "naive_date",
    "begin": "2020-01-01",
    "end": "2025-01-01"
  }
}
```

Accepted values for the `"date_time"` key are objects with the following keys:

- `"format"`: a [strftime](https://docs.rs/chrono/0.4.19/chrono/format/strftime/index.html)-style parameter specifying
  the string formatting of the underlying `date_time`
  value.
- `"subtype"`: one of the following
    * `"naive_date"`: when the underlying `date_time` value should be a simple `date` without timezone specification,
    * `"naive_time"`: when the underlying `date_time` value should be a simple `time` without timezone specification,
    * `"naive_date_time"`: when the underlying `date_time` value should be a combined `date` and `time` without timezone
      specification,
    * `"date_time"`: when the underlying `date_time` value should be a combined `date` and `time` *with* timezone
      specification.
- `"begin"` and `"end"`: the lower and upper bounds of the `date_time` value to generate. The formatting of these values
  must adhere to the `strftime`-string specified in the `"format"` field.

#### Example

```json synth
{
  "type": "string",
  "date_time": {
    "format": "%Y-%m-%dT%H:%M:%S",
    "subtype": "naive_date_time",
    "begin": "2015-01-01T00:00:00",
    "end": "2020-01-01T12:00:00"
  }
}
```

## serialized

`serialized` is a variant of the `string` generator type which serializes the output of a child generator into a string.

`serialized` has 2 fields,
- `serializer`: The serializer to be used (currently only `json`)
- `content`: The content to be serialized. This can be any valid Synth generator



#### Example

```json synth
{
    "type": "string",
    "serialized": {
        "serializer": "json",
        "content": {
            "type": "object",
            "username": {
                "type": "string",
                "faker": {
                    "generator": "name"
                }
            }
        }
    }
}
```

## truncated

The `truncated` generator truncates the output of it's inner generator to a fixed length.

If the output of its inner generator is less than or equal to the length, it is left untouched. 

`truncated` has 2 fields,
- `length`: The number of characters to truncate to.
- `content`: The content to be truncated. This can be any Synth generator that yields a String.

#### Example

```json synth
{
  "type": "string",
  "truncated": {
    "content": {
      "type": "string",
      "pattern": "[a-zA-Z0-9]{0, 255}"
    },
    "length": 5
  }
}
```



## categorical

A `categorical` is a variant of the `string` generator type that generates values from a finite set of user-defined
values. You can specify a string as a categorical by using the `"categorical": {...}` key/value pair.

#### Example

```json synth
{
  "type": "string",
  "categorical": {
    "pawn": 8,
    "rook": 2,
    "knight": 2,
    "bishop": 2,
    "queen": 1,
    "king": 1
  }
}
```

The value of the `"categorical"` key must be an object whose:

- keys are the allowed values of the categorical (e.g. `"pawn"`, `"rook"`, etc.),
- values are non-negative integers defining the relative weight of the corresponding variant (e.g. `8`, `2`, etc.).

[faker]: https://faker.readthedocs.io/en/master/